import { DataGrid } from 'tde-datagrid';
import { addIcon } from '@jupyterlab/ui-components';
export class PaintedGrid extends DataGrid {
    constructor(options) {
        super(options);
        this._extraStyle = options.extraStyle || Private.defaultExtraStyle;
    }
    /**
     * Get the extra styles of the PaintedGrid.
     */
    get extraStyle() {
        return this._extraStyle;
    }
    /**
     * Set the styles of the painted grid.
     */
    set extraStyle(value) {
        // Bail if the style does not change.
        if (this._extraStyle === value) {
            return;
        }
        // Update the internal style.
        this._extraStyle = Object.assign({}, value);
        // Schedule a repaint of the content.
        this.repaintContent();
        // Schedule a repaint of the overlay.
        this.repaintOverlay();
    }
    /**
     * @override paints on the ghost row and column as well after painting the other regions.
     * Paint the grid content for the given dirty rect.
     *
     * The rect should be expressed in valid viewport coordinates.
     *
     * This is the primary paint entry point. The individual `_draw*`
     * methods should not be invoked directly. This method dispatches
     * to the drawing methods in the correct order.
     */
    paintContent(rx, ry, rw, rh) {
        // Paint in the background, rows, columns, and cells first.
        super.paintContent(rx, ry, rw, rh);
        // Paint addons.
        this._paintAddons(rx, ry, rw, rh);
    }
    /**
     * Primary entry point for painting additional graphics on top of
     * the base data grid graphics.
     */
    _paintAddons(rx, ry, rw, rh) {
        // Draw the ghost row.
        this._drawGhostRow(rx, ry, rw, rh);
        // Draw the header region for the ghost row.
        this._drawGhostRowHeader(rx, ry, rw, rh);
        // Draw the ghost column.
        this._drawGhostColumn(rx, ry, rw, rh);
        // Draw the header region for the ghost column.
        this._drawGhostColumnHeader(rx, ry, rw, rh);
        // Draw over the corner to hide it from view.
        this._drawOverCorner(rx, ry, rw, rh);
    }
    /**
     * Draw the ghost row.
     */
    _drawGhostRow(rx, ry, rw, rh) {
        // Get the visible content dimensions.
        const contentW = this.bodyWidth - this.scrollX;
        const contentH = this.defaultSizes.rowHeight;
        // Bail if there is no content to draw.
        if (contentW <= 0 || contentH <= 0) {
            return;
        }
        // Get the visible content origin.
        const contentX = this.headerWidth;
        const contentY = this.headerHeight + this.bodyHeight - contentH - this.scrollY;
        // Bail if the dirty rect does not intersect the content area.
        if (rx + rw <= contentX) {
            return;
        }
        if (ry + rh <= contentY) {
            return;
        }
        if (rx >= contentX + contentW) {
            return;
        }
        if (ry >= contentY + contentH) {
            return;
        }
        // Get the upper and lower bounds of the dirty content area.
        const x1 = Math.max(rx, contentX);
        const y1 = Math.max(ry, contentY);
        const x2 = Math.min(rx + rw - 1, contentX + contentW - 1);
        const y2 = Math.min(ry + rh - 1, contentY + contentH - 1);
        // Fill the region with the specified color.
        this.canvasGC.fillStyle = this._extraStyle.ghostRowColor;
        this.canvasGC.fillRect(x1, y1, x2 - x1 + 1, y2 - y1 + 1);
    }
    /**
     * Draw the ghost column.
     */
    _drawGhostColumn(rx, ry, rw, rh) {
        // Get the visible content dimensions.
        const contentW = this.defaultSizes.columnWidth;
        const contentH = this.bodyHeight - this.scrollY;
        // Bail if there is no content to draw.
        if (contentW <= 0 || contentH <= 0) {
            return;
        }
        // Get the visible content origin.
        const contentX = this.headerWidth + this.bodyWidth - contentW - this.scrollX;
        const contentY = this.headerHeight;
        // Bail if the dirty rect does not intersect the content area.
        if (rx + rw <= contentX) {
            return;
        }
        if (ry + rh <= contentY) {
            return;
        }
        if (rx >= contentX + contentW) {
            return;
        }
        if (ry >= contentY + contentH) {
            return;
        }
        // Get the upper and lower bounds of the dirty content area.
        const x1 = Math.max(rx, contentX);
        const y1 = Math.max(ry, contentY);
        const x2 = Math.min(rx + rw - 1, contentX + contentW - 1);
        const y2 = Math.min(ry + rh - 1, contentY + contentH - 1);
        // Fill the region with the specified color.
        this.canvasGC.fillStyle = this._extraStyle.ghostColumnColor;
        this.canvasGC.fillRect(x1, y1, x2 - x1 + 1, y2 - y1 + 1);
    }
    _drawGhostRowHeader(rx, ry, rw, rh) {
        // Get the visible content dimensions.
        const contentW = this.headerWidth;
        const contentH = this.defaultSizes.rowHeight;
        // Bail if there is no content to draw.
        if (contentW <= 0 || contentH <= 0) {
            return;
        }
        // Get the visible content origin.
        const contentX = 0;
        const contentY = this.headerHeight + this.bodyHeight - contentH - this.scrollY;
        // Bail if the dirty rect does not intersect the content area.
        if (rx + rw <= contentX) {
            return;
        }
        if (ry + rh <= contentY) {
            return;
        }
        if (rx >= contentX + contentW) {
            return;
        }
        if (ry >= contentY + contentH) {
            return;
        }
        // Get the upper and lower bounds of the dirty content area.
        const x1 = rx;
        const y1 = Math.max(ry, contentY);
        const x2 = Math.min(rx + rw - 1, contentX + contentW - 1);
        const y2 = Math.min(ry + rh - 1, contentY + contentH - 1);
        // Fill the region with the specified color.
        this.canvasGC.fillStyle = this._extraStyle.ghostColumnColor;
        this.canvasGC.fillRect(x1, y1, x2 - x1 + 1, y2 - y1 + 1);
        this._drawGhostRowIcon();
    }
    _drawGhostColumnHeader(rx, ry, rw, rh) {
        // Get the visible content dimensions.
        const contentW = this.defaultSizes.columnWidth;
        const contentH = this.headerHeight;
        // Bail if there is no content to draw.
        if (contentW <= 0 || contentH <= 0) {
            return;
        }
        // Get the visible content origin.
        const contentX = this.headerWidth + this.bodyWidth - contentW - this.scrollX;
        const contentY = 0;
        // Bail if the dirty rect does not intersect the content area.
        if (rx + rw <= contentX) {
            return;
        }
        if (ry + rh <= contentY) {
            return;
        }
        if (rx >= contentX + contentW) {
            return;
        }
        if (ry >= contentY + contentH) {
            return;
        }
        // Get the upper and lower bounds of the dirty content area.
        const x1 = Math.max(rx, contentX);
        const y1 = Math.max(ry, contentY);
        const x2 = Math.min(rx + rw - 1, contentX + contentW - 1);
        const y2 = Math.min(ry + rh - 1, contentY + contentH - 1);
        // Fill the region with the specified color.
        this.canvasGC.fillStyle = this._extraStyle.ghostColumnColor;
        this.canvasGC.fillRect(x1, y1, x2 - x1 + 1, y2 - y1 + 1);
        this._drawColumnIcon();
    }
    _drawGhostRowIcon() {
        // Get the dimensions for the cell.
        const cellW = this.headerWidth;
        const cellH = this.defaultSizes.rowHeight;
        // Get the icon arguments.
        const iconArgs = this._extraStyle.icons['ghost-row'];
        // Bail early if there are no icon arguments.
        if (!iconArgs) {
            return;
        }
        // Get the current transform state.
        const transform = this.canvasGC.getTransform();
        // Unpack the icon arguments.
        const { icon, color, height } = iconArgs;
        // Parse the icon path from the icon string.
        const { defaultHeight, viewBoxSize, path } = Private.parseSVG(icon.svgstr);
        // Create a path 2d object from the path string.
        const canvasPath = new Path2D(path);
        // Solve for the scaling factor using the provided width or the default.
        const scale = (height * Math.min(cellW, cellH)) / defaultHeight;
        // Orient the canvas to the desired origin for the icon.
        this.canvasGC.translate(cellW / 2 - (viewBoxSize * scale) / 2, this.headerHeight +
            this.bodyHeight -
            this.scrollY -
            cellH / 2 -
            (viewBoxSize * scale) / 2);
        // Scale the canvas.
        this.canvasGC.scale(scale, scale);
        // Set the canvas fill style.
        this.canvasGC.fillStyle = color;
        // Draw the icon.
        this.canvasGC.fill(canvasPath, 'nonzero');
        // Reset the canvas to it's default position.
        this.canvasGC.setTransform(transform);
    }
    _drawColumnIcon() {
        // Get the dimensions for the cell.
        const cellW = this.defaultSizes.columnWidth;
        const cellH = this.headerHeight;
        // Get the icon arguments.
        const iconArgs = this._extraStyle.icons['ghost-column'];
        // Bail early if there are no icon arguments.
        if (!iconArgs) {
            return;
        }
        // Get the current transform state.
        const transform = this.canvasGC.getTransform();
        // Unpack the icon arguments.
        const { icon, color, height, left, top } = iconArgs;
        // Parse the icon path from the icon string.
        const { defaultHeight, viewBoxSize, path } = Private.parseSVG(icon.svgstr);
        // Create a path 2d object from the path string.
        const canvasPath = new Path2D(path);
        // Solve for the scaling factor using the provided width or the default.
        const scale = (height * Math.min(cellH, cellW)) / defaultHeight;
        // Orient to the desired origin for the icon.
        this.canvasGC.translate(this.headerWidth +
            this.bodyWidth -
            this.scrollX -
            left * cellW -
            (viewBoxSize * scale) / 2, top * cellH - (viewBoxSize * scale) / 2);
        // Scale the canvas.
        this.canvasGC.scale(scale, scale);
        // Set the canvas fill style.
        this.canvasGC.fillStyle = color;
        // Draw the icon.
        this.canvasGC.fill(canvasPath, 'nonzero');
        // Reset the transform to the initial state
        this.canvasGC.setTransform(transform);
    }
    _drawOverCorner(rx, ry, rw, rh) {
        // Get the visible content dimensions.
        const contentW = this.defaultSizes.columnWidth;
        const contentH = this.defaultSizes.rowHeight;
        // Bail if there is no content to draw.
        if (contentW <= 0 || contentH <= 0) {
            return;
        }
        // Get the visible content origin.
        const contentX = this.headerWidth + this.bodyWidth - contentW - this.scrollX;
        const contentY = this.headerHeight + this.bodyHeight - contentH - this.scrollY;
        // Bail if the dirty rect does not intersect the content area.
        if (rx + rw <= contentX) {
            return;
        }
        if (ry + rh <= contentY) {
            return;
        }
        if (rx >= contentX + contentW) {
            return;
        }
        if (ry >= contentY + contentH) {
            return;
        }
        // Get the upper and lower bounds of the dirty content area.
        const x1 = Math.max(rx, contentX);
        const y1 = Math.max(ry, contentY);
        const x2 = Math.min(rx + rw - 1, contentX + contentW - 1);
        const y2 = Math.min(ry + rh - 1, contentY + contentH - 1);
        // Fill the region with the specified color.
        this.canvasGC.fillStyle = this.style.voidColor;
        this.canvasGC.fillRect(x1, y1, x2 - x1 + 1, y2 - y1 + 1);
    }
}
/**
 * Namespace for module implementation details.
 */
var Private;
(function (Private) {
    /**
     * Parse an svg string into a standard form.
     */
    function parseSVG(svgstr) {
        // Set up a regular expression to get the width.
        let regex = /width="(.+?)"/;
        // Find the width an parse it to an integer.
        const defaultHeight = parseInt(svgstr.match(regex)[1]);
        // Redefine the regular expression to get the viewbox size.
        regex = /viewBox="(.+?)"/;
        const viewBox = svgstr
            .match(regex)[1]
            .split(' ')
            .map(digit => parseInt(digit));
        const viewBoxSize = viewBox[2];
        // Redefine the regular expression to get the path string.
        regex = /path d="(.+?)"/;
        // Fetch the path string.
        const path = svgstr.match(regex)[1];
        return { defaultHeight, viewBoxSize, path };
    }
    Private.parseSVG = parseSVG;
    Private.defaultExtraStyle = {
        ghostRowColor: 'rgba(243, 243, 243, 0.80)',
        ghostColumnColor: 'rgba(243, 243, 243, 0.80)',
        icons: {
            'ghost-column': {
                icon: addIcon,
                color: '#616161',
                height: 1 / 2,
                left: 1 / 2,
                top: 1 / 2
            },
            'ghost-row': {
                icon: addIcon,
                color: '#616161',
                height: 1 / 2,
                left: 1 / 2,
                top: 1 / 2
            }
        }
    };
})(Private || (Private = {}));
//# sourceMappingURL=grid.js.map